@@ -10,7 +10,6 @@
 
 struct rpmdsObject_s {
     PyObject_HEAD
-    PyObject *md_dict;		/*!< to look like PyModuleObject */
     int		active;
     rpmds	ds;
 };
@@ -82,6 +81,12 @@ rpmds_IsReverse(rpmdsObject * s)
     return PyBool_FromLong(rpmdsIsReverse(s->ds));
 }
 
+static PyObject *
+rpmds_IsSysuser(rpmdsObject * s)
+{
+    return PyBool_FromLong(rpmdsIsSysuser(s->ds, NULL));
+}
+
 static PyObject *
 rpmds_iternext(rpmdsObject * s)
 {
@@ -102,19 +107,6 @@ rpmds_iternext(rpmdsObject * s)
     return result;
 }
 
-static PyObject *
-rpmds_SetNoPromote(rpmdsObject * s, PyObject * args, PyObject * kwds)
-{
-    int nopromote;
-    char * kwlist[] = {"noPromote", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "i:SetNoPromote", kwlist,
-	    &nopromote))
-	return NULL;
-
-    return Py_BuildValue("i", rpmdsSetNoPromote(s->ds, nopromote));
-}
-
 /* XXX rpmdsFind uses bsearch on s->ds, so a sort is needed. */
 static PyObject *
 rpmds_Sort(rpmdsObject * s)
@@ -127,8 +119,12 @@ static PyObject *
 rpmds_Find(rpmdsObject * s, PyObject * arg)
 {
     rpmdsObject * o;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
 
-    if (!PyArg_Parse(arg, "O!:Find", &rpmds_Type, &o))
+    if (!PyArg_Parse(arg, "O!:Find", modstate->rpmds_Type, &o))
 	return NULL;
 
     /* XXX make sure ods index is valid, real fix in lib/rpmds.c. */
@@ -141,8 +137,12 @@ static PyObject *
 rpmds_Merge(rpmdsObject * s, PyObject * arg)
 {
     rpmdsObject * o;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
 
-    if (!PyArg_Parse(arg, "O!:Merge", &rpmds_Type, &o))
+    if (!PyArg_Parse(arg, "O!:Merge", modstate->rpmds_Type, &o))
 	return NULL;
 
     return Py_BuildValue("i", rpmdsMerge(&s->ds, o->ds));
@@ -151,8 +151,12 @@ static PyObject *
 rpmds_Search(rpmdsObject * s, PyObject * arg)
 {
     rpmdsObject * o;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
 
-    if (!PyArg_Parse(arg, "O!:Merge", &rpmds_Type, &o))
+    if (!PyArg_Parse(arg, "O!:Merge", modstate->rpmds_Type, &o))
         return NULL;
 
     return Py_BuildValue("i", rpmdsSearch(s->ds, o->ds));
@@ -161,8 +165,12 @@ rpmds_Search(rpmdsObject * s, PyObject * arg)
 static PyObject *rpmds_Compare(rpmdsObject * s, PyObject * o)
 {
     rpmdsObject * ods;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
 
-    if (!PyArg_Parse(o, "O!:Compare", &rpmds_Type, &ods))
+    if (!PyArg_Parse(o, "O!:Compare", modstate->rpmds_Type, &ods))
 	return NULL;
 
     return PyBool_FromLong(rpmdsCompare(s->ds, ods->ds));
@@ -175,18 +183,27 @@ static PyObject *rpmds_Instance(rpmdsObject * s)
 
 static PyObject * rpmds_Rpmlib(rpmdsObject * s, PyObject *args, PyObject *kwds)
 {
+    PyObject *pool_source = NULL;
     rpmstrPool pool = NULL;
     rpmds ds = NULL;
     char * kwlist[] = {"pool", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&:rpmds_Rpmlib", kwlist, 
-		 &poolFromPyObject, &pool))
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
 	return NULL;
+    }
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O:rpmds_Rpmlib", kwlist,
+		 &pool_source))
+	return NULL;
+
+    if(pool_source && !poolFromPyObject(modstate, pool_source, &pool)) {
+	    return NULL;
+    }
 
     /* XXX check return code, permit arg (NULL uses system default). */
     rpmdsRpmlibPool(pool, &ds, NULL);
 
-    return rpmds_Wrap(&rpmds_Type, ds);
+    return rpmds_Wrap(modstate->rpmds_Type, ds);
 }
 
 static struct PyMethodDef rpmds_methods[] = {
@@ -207,9 +224,6 @@ static struct PyMethodDef rpmds_methods[] = {
   "the type of all dependencies in this set.\n" },
  {"Color",	(PyCFunction)rpmds_Color,	METH_NOARGS,
 	"ds.Color -> Color -- Return current Color.\n" },
- {"SetNoPromote",(PyCFunction)rpmds_SetNoPromote, METH_VARARGS|METH_KEYWORDS,
-  "ds.SetNoPromote(noPromote) -- Set noPromote for this instance.\n\n"
-  "If True non existing epochs are no longer equal to an epoch of 0."},
  {"Sort",	(PyCFunction)rpmds_Sort,	METH_NOARGS,
 	NULL},
  {"Find",	(PyCFunction)rpmds_Find,	METH_O,
@@ -232,6 +246,8 @@ The current index in ds is positioned at overlapping member." },
   "ds.IsRich() -- Return whether the dependency is rich."},
  {"IsReverse",	(PyCFunction)rpmds_IsReverse,	METH_NOARGS,
   "ds.IsReverse() -- Return whether the dependency is reversed."},
+ {"IsSysuser",	(PyCFunction)rpmds_IsSysuser,	METH_NOARGS,
+  "ds.IsSysuser() -- Return whether the dependency represents a sysusers.d entry."},
  {NULL,		NULL}		/* sentinel */
 };
 
@@ -240,8 +256,20 @@ The current index in ds is positioned at overlapping member." },
 static void
 rpmds_dealloc(rpmdsObject * s)
 {
+    PyObject_GC_UnTrack(s);
     s->ds = rpmdsFree(s->ds);
-    Py_TYPE(s)->tp_free((PyObject *)s);
+    PyTypeObject *type = Py_TYPE(s);
+    freefunc free = PyType_GetSlot(type, Py_tp_free);
+    free(s);
+    Py_DECREF(type);
+}
+
+static int rpmds_traverse(hdrObject * s, visitproc visit, void *arg)
+{
+    if (python_version >= 0x03090000) {
+        Py_VISIT(Py_TYPE(s));
+    }
+    return 0;
 }
 
 static Py_ssize_t rpmds_length(rpmdsObject * s)
@@ -254,21 +282,16 @@ rpmds_subscript(rpmdsObject * s, PyObject * key)
 {
     int ix;
 
-    if (!PyInt_Check(key)) {
+    if (!PyLong_Check(key)) {
 	PyErr_SetString(PyExc_TypeError, "integer expected");
 	return NULL;
     }
 
-    ix = (int) PyInt_AsLong(key);
+    ix = (int) PyLong_AsLong(key);
     rpmdsSetIx(s->ds, ix);
     return utf8FromString(rpmdsDNEVR(s->ds));
 }
 
-static PyMappingMethods rpmds_as_mapping = {
-        (lenfunc) rpmds_length,		/* mp_length */
-        (binaryfunc) rpmds_subscript,	/* mp_subscript */
-        (objobjargproc)0,		/* mp_ass_subscript */
-};
 
 static int rpmds_init(rpmdsObject * s, PyObject *args, PyObject *kwds)
 {
@@ -282,9 +305,9 @@ static int depflags(PyObject *o, rpmsenseFlags *senseFlags)
     PyObject *str = NULL;
     rpmsenseFlags flags = RPMSENSE_ANY;
 
-    if (PyInt_Check(o)) {
+    if (PyLong_Check(o)) {
 	ok = 1;
-	flags = PyInt_AsLong(o);
+	flags = PyLong_AsLong(o);
     } else if (utf8FromPyObject(o, &str)) {
 	ok = 1;
 	for (const char *s = PyBytes_AsString(str); *s; s++) {
@@ -321,13 +344,22 @@ static PyObject * rpmds_new(PyTypeObject * subtype, PyObject *args, PyObject *kw
     rpmTagVal tagN = RPMTAG_REQUIRENAME;
     rpmds ds = NULL;
     Header h = NULL;
+    PyObject *pool_source = NULL;
     rpmstrPool pool = NULL;
     char * kwlist[] = {"obj", "tag", "pool", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "OO&|O&:rpmds_new", kwlist, 
-	    	 &obj, tagNumFromPyObject, &tagN,
-		 &poolFromPyObject, &pool))
+    rpmmodule_state_t *modstate = rpmModState_FromType(subtype);
+    if (!modstate) {
 	return NULL;
+    }
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "OO&|O:rpmds_new", kwlist,
+	    	 &obj, tagNumFromPyObject, &tagN,
+		 &pool_source))
+	return NULL;
+
+    if(pool_source && !poolFromPyObject(modstate, pool_source, &pool)) {
+	return NULL;
+    }
 
     if (PyTuple_Check(obj)) {
 	const char *name = NULL;
@@ -359,50 +391,35 @@ static char rpmds_doc[] =
     "It can hold multiple entries of Name Flags and EVR.\n"
     "It typically represents all dependencies of one kind of a package\n"
     "e.g. all Requires or all Conflicts.\n"
+    "\n"
+    "A dependency set can be instantiated from a header or a tuple:\n"
+    "ds(hdr, tag, pool=None)\n"
+    "ds((e, v, r), tag, pool=None)\n\n"
+    "tag is the dependency type, eg 'requires', 'provides', 'obsoletes' etc"
+    "or the corresponding RPMTAG_* value. As a special case, 'nevr' returns"
+    "a provide of the header NEVR."
     ;
 
-PyTypeObject rpmds_Type = {
-	PyVarObject_HEAD_INIT(&PyType_Type, 0)
-	"rpm.ds",			/* tp_name */
-	sizeof(rpmdsObject),		/* tp_basicsize */
-	0,				/* tp_itemsize */
-	/* methods */
-	(destructor) rpmds_dealloc,	/* tp_dealloc */
-	0,				/* tp_print */
-	(getattrfunc)0,			/* tp_getattr */
-	(setattrfunc)0,			/* tp_setattr */
-	0,				/* tp_compare */
-	(reprfunc)0,			/* tp_repr */
-	0,				/* tp_as_number */
-	0,				/* tp_as_sequence */
-	&rpmds_as_mapping,		/* tp_as_mapping */
-	(hashfunc)0,			/* tp_hash */
-	(ternaryfunc)0,			/* tp_call */
-	(reprfunc)0,			/* tp_str */
-	PyObject_GenericGetAttr,	/* tp_getattro */
-	PyObject_GenericSetAttr,	/* tp_setattro */
-	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
-	rpmds_doc,			/* tp_doc */
-	0,				/* tp_traverse */
-	0,				/* tp_clear */
-	0,				/* tp_richcompare */
-	0,				/* tp_weaklistoffset */
-	PyObject_SelfIter,		/* tp_iter */
-	(iternextfunc) rpmds_iternext,	/* tp_iternext */
-	rpmds_methods,			/* tp_methods */
-	0,				/* tp_members */
-	0,				/* tp_getset */
-	0,				/* tp_base */
-	0,				/* tp_dict */
-	0,				/* tp_descr_get */
-	0,				/* tp_descr_set */
-	0,				/* tp_dictoffset */
-	(initproc) rpmds_init,		/* tp_init */
-	0,				/* tp_alloc */
-	(newfunc) rpmds_new,		/* tp_new */
-	0,				/* tp_free */
-	0,				/* tp_is_gc */
+static PyType_Slot rpmds_Type_Slots[] = {
+    {Py_tp_dealloc, rpmds_dealloc},
+    {Py_tp_traverse, rpmds_traverse},
+    {Py_mp_length, rpmds_length},
+    {Py_mp_subscript, rpmds_subscript},
+    {Py_tp_getattro, PyObject_GenericGetAttr},
+    {Py_tp_setattro, PyObject_GenericSetAttr},
+    {Py_tp_doc, rpmds_doc},
+    {Py_tp_iter, PyObject_SelfIter},
+    {Py_tp_iternext, rpmds_iternext},
+    {Py_tp_methods, rpmds_methods},
+    {Py_tp_init, rpmds_init},
+    {Py_tp_new, rpmds_new},
+    {0, NULL},
+};
+PyType_Spec rpmds_Type_Spec = {
+    .name = "rpm.ds",
+    .basicsize = sizeof(rpmdsObject),
+    .flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_IMMUTABLETYPE,
+    .slots = rpmds_Type_Slots,
 };
 
 /* ---------- */
@@ -414,7 +431,8 @@ rpmds dsFromDs(rpmdsObject * s)
 
 PyObject * rpmds_Wrap(PyTypeObject *subtype, rpmds ds)
 {
-    rpmdsObject * s = (rpmdsObject *)subtype->tp_alloc(subtype, 0);
+    allocfunc subtype_alloc = (allocfunc)PyType_GetSlot(subtype, Py_tp_alloc);
+    rpmdsObject *s = (rpmdsObject *)subtype_alloc(subtype, 0);
     if (s == NULL) return NULL;
 
     s->ds = ds;
