@@ -29,7 +29,7 @@
  *	import rpm
  *	ts = rpm.TransactionSet()
  *	for h in ts.dbMatch():
- *	    print h['name']
+ *	    print(h['name'])
  * \endcode
  *
  * Here's a more typical example that uses the Name index to retrieve
@@ -39,7 +39,7 @@
  *	ts = rpm.TransactionSet()
  *	mi = ts.dbMatch('name', 'kernel')
  *	for h in mi:
- *	    print '%s-%s-%s' % (h['name'], h['version'], h['release'])
+ *	    print('%s-%s-%s' % (h['name'], h['version'], h['release']))
  * \endcode
  *
  * Finally, here's an example that retrieves all packages whose name
@@ -50,7 +50,7 @@
  *	mi = ts.dbMatch()
  *	mi.pattern('name', rpm.RPMMIRE_GLOB, 'XFree*')
  *	for h in mi:
- *	    print '%s-%s-%s' % (h['name'], h['version'], h['release'])
+ *	    print('%s-%s-%s' % (h['name'], h['version'], h['release']))
  * \endcode
  *
  */
@@ -61,7 +61,6 @@
 
 struct rpmmiObject_s {
     PyObject_HEAD
-    PyObject *md_dict;		/*!< to look like PyModuleObject */
     PyObject *ref;		/* for db/ts refcounting */
     rpmdbMatchIterator mi;
 } ;
@@ -70,13 +69,17 @@ static PyObject *
 rpmmi_iternext(rpmmiObject * s)
 {
     Header h;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+        return NULL;
+    }
 
     if (s->mi == NULL || (h = rpmdbNextIterator(s->mi)) == NULL) {
 	s->mi = rpmdbFreeIterator(s->mi);
 	return NULL;
     }
     headerLink(h);
-    return hdr_Wrap(&hdr_Type, h);
+    return hdr_Wrap(modstate->hdr_Type, h);
 }
 
 static PyObject *
@@ -127,9 +130,22 @@ static struct PyMethodDef rpmmi_methods[] = {
 
 static void rpmmi_dealloc(rpmmiObject * s)
 {
+    PyObject_GC_UnTrack(s);
     s->mi = rpmdbFreeIterator(s->mi);
     Py_DECREF(s->ref);
-    Py_TYPE(s)->tp_free((PyObject *)s);
+    PyTypeObject *type = Py_TYPE(s);
+    freefunc free = PyType_GetSlot(type, Py_tp_free);
+    free(s);
+    Py_DECREF(type);
+}
+
+static int rpmmi_traverse(rpmmiObject * s, visitproc visit, void *arg)
+{
+    Py_VISIT(s->ref);
+    if (python_version >= 0x03090000) {
+        Py_VISIT(Py_TYPE(s));
+    }
+    return 0;
 }
 
 static Py_ssize_t rpmmi_length(rpmmiObject * s)
@@ -142,26 +158,7 @@ static int rpmmi_bool(rpmmiObject *s)
     return (s->mi != NULL);
 }
 
-PyMappingMethods rpmmi_as_mapping = {
-    (lenfunc) rpmmi_length,		/* mp_length */
-    0,
-};
 
-static PyNumberMethods rpmmi_as_number = {
-	0, /* nb_add */
-	0, /* nb_subtract */
-	0, /* nb_multiply */
-#if PY_MAJOR_VERSION < 3
-	0, /* nb_divide */
-#endif
-	0, /* nb_remainder */
-	0, /* nb_divmod */
-	0, /* nb_power */
-	0, /* nb_negative */
-	0, /* nb_positive */
-	0, /* nb_absolute */
-	(inquiry)rpmmi_bool, /* nb_bool/nonzero */
-};
 
 static char rpmmi_doc[] =
   "rpm.mi match iterator object represents the result of a\n"
@@ -179,7 +176,7 @@ static char rpmmi_doc[] =
   "	import rpm\n"
   "	ts = rpm.TransactionSet()\n"
   "	for h in ts.dbMatch():\n"
-  "	    print h['name']\n"
+  "	    print(h['name'])\n"
   "\n"
   "Here's a more typical example that uses the Name index to retrieve\n"
   "all installed kernel(s):\n"
@@ -187,7 +184,7 @@ static char rpmmi_doc[] =
   "	ts = rpm.TransactionSet()\n"
   "	mi = ts.dbMatch('name', 'kernel')\n"
   "	for h in mi:\n"
-  "	    print '%s-%s-%s' % (h['name'], h['version'], h['release'])\n"
+  "	    print('%s-%s-%s' % (h['name'], h['version'], h['release']))\n"
   "\n"
   "Finally, here's an example that retrieves all packages whose name\n"
   "matches the glob expression 'XFree*':\n"
@@ -196,55 +193,42 @@ static char rpmmi_doc[] =
   "	mi = ts.dbMatch()\n"
   "	mi.pattern('name', rpm.RPMMIRE_GLOB, 'XFree*')\n"
   "	for h in mi:\n"
-  "	    print '%s-%s-%s' % (h['name'], h['version'], h['release'])\n"
+  "	    print('%s-%s-%s' % (h['name'], h['version'], h['release']))\n"
 ;
 
-PyTypeObject rpmmi_Type = {
-	PyVarObject_HEAD_INIT(&PyType_Type, 0)
-	"rpm.mi",			/* tp_name */
-	sizeof(rpmmiObject),		/* tp_size */
-	0,				/* tp_itemsize */
-	(destructor) rpmmi_dealloc, 	/* tp_dealloc */
-	0,				/* tp_print */
-	(getattrfunc)0, 		/* tp_getattr */
-	0,				/* tp_setattr */
-	0,				/* tp_compare */
-	0,				/* tp_repr */
-	&rpmmi_as_number,		/* tp_as_number */
-	0,				/* tp_as_sequence */
-	&rpmmi_as_mapping,		/* tp_as_mapping */
-	0,				/* tp_hash */
-	0,				/* tp_call */
-	0,				/* tp_str */
-	PyObject_GenericGetAttr,	/* tp_getattro */
-	PyObject_GenericSetAttr,	/* tp_setattro */
-	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,	/* tp_flags */
-	rpmmi_doc,			/* tp_doc */
-	0,				/* tp_traverse */
-	0,				/* tp_clear */
-	0,				/* tp_richcompare */
-	0,				/* tp_weaklistoffset */
-	PyObject_SelfIter,		/* tp_iter */
-	(iternextfunc) rpmmi_iternext,	/* tp_iternext */
-	rpmmi_methods,			/* tp_methods */
-	0,				/* tp_members */
-	0,				/* tp_getset */
-	0,				/* tp_base */
-	0,				/* tp_dict */
-	0,				/* tp_descr_get */
-	0,				/* tp_descr_set */
-	0,				/* tp_dictoffset */
-	0,				/* tp_init */
-	0,				/* tp_alloc */
-	0,				/* tp_new */
-	0,				/* tp_free */
-	0,				/* tp_is_gc */
+static PyObject *disabled_new(PyTypeObject *type,
+                              PyObject *args, PyObject *kwds)
+{
+    PyErr_SetString(PyExc_TypeError,
+                    "TypeError: cannot create 'rpm.mi' instances");
+    return NULL;
+}
+
+static PyType_Slot rpmmi_Type_Slots[] = {
+    {Py_tp_new, disabled_new},
+    {Py_tp_dealloc, rpmmi_dealloc},
+    {Py_tp_traverse, rpmmi_traverse},
+    {Py_nb_bool, rpmmi_bool},
+    {Py_mp_length, rpmmi_length},
+    {Py_tp_getattro, PyObject_GenericGetAttr},
+    {Py_tp_setattro, PyObject_GenericSetAttr},
+    {Py_tp_doc, rpmmi_doc},
+    {Py_tp_iter, PyObject_SelfIter},
+    {Py_tp_iternext, rpmmi_iternext},
+    {Py_tp_methods, rpmmi_methods},
+    {0, NULL},
+};
+PyType_Spec rpmmi_Type_Spec = {
+    .name = "rpm.mi",
+    .basicsize = sizeof(rpmmiObject),
+    .flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_IMMUTABLETYPE,
+    .slots = rpmmi_Type_Slots,
 };
 
 PyObject * rpmmi_Wrap(PyTypeObject *subtype, rpmdbMatchIterator mi, PyObject *s)
 {
-    rpmmiObject * mio = (rpmmiObject *)subtype->tp_alloc(subtype, 0);
+    allocfunc subtype_alloc = (allocfunc)PyType_GetSlot(subtype, Py_tp_alloc);
+    rpmmiObject *mio = (rpmmiObject *)subtype_alloc(subtype, 0);
     if (mio == NULL) return NULL;
 
     mio->mi = mi;
