@@ -4,7 +4,6 @@
 
 #include <rpm/rpmlib.h>	/* rpmReadPackageFile, headerCheck */
 #include <rpm/rpmtag.h>
-#include <rpm/rpmpgp.h>
 #include <rpm/rpmdb.h>
 #include <rpm/rpmbuild.h>
 
@@ -12,7 +11,6 @@
 #include "rpmds-py.h"	/* XXX for rpmdsNew */
 #include "rpmfd-py.h"
 #include "rpmkeyring-py.h"
-#include "rpmfi-py.h"	/* XXX for rpmfiNew */
 #include "rpmmi-py.h"
 #include "rpmii-py.h"
 #include "rpmps-py.h"
@@ -139,7 +137,6 @@
 
 struct rpmtsObject_s {
     PyObject_HEAD
-    PyObject *md_dict;		/*!< to look like PyModuleObject */
     rpmfdObject *scriptFd;
     PyObject *keyList;
     rpmts	ts;
@@ -150,6 +147,7 @@ struct rpmtsCallbackType_s {
     PyObject * cb;
     PyObject * data;
     rpmtsObject * tso;
+    int style;
     PyThreadState *_save;
 };
 
@@ -229,6 +227,17 @@ rpmts_AddErase(rpmtsObject * s, PyObject * args)
     return PyBool_FromLong(rpmtsAddEraseElement(s->ts, h, -1) == 0);
 }
 
+static PyObject *
+rpmts_AddRestore(rpmtsObject * s, PyObject * args)
+{
+    Header h;
+
+    if (!PyArg_ParseTuple(args, "O&:AddRestore", hdrFromPyObject, &h))
+        return NULL;
+
+    return PyBool_FromLong(rpmtsAddRestoreElement(s->ts, h) == 0);
+}
+
 static int
 rpmts_SolveCallback(rpmts ts, rpmds ds, const void * data)
 {
@@ -244,14 +253,14 @@ rpmts_SolveCallback(rpmts ts, rpmds ds, const void * data)
     args = Py_BuildValue("(OiNNi)", cbInfo->tso,
 		rpmdsTagN(ds), utf8FromString(rpmdsN(ds)),
 		utf8FromString(rpmdsEVR(ds)), rpmdsFlags(ds));
-    result = PyEval_CallObject(cbInfo->cb, args);
+    result = PyObject_Call(cbInfo->cb, args, NULL);
     Py_DECREF(args);
 
     if (!result) {
 	die(cbInfo->cb);
     } else {
-	if (PyInt_Check(result))
-	    res = PyInt_AsLong(result);
+	if (PyLong_Check(result))
+	    res = PyLong_AsLong(result);
 	Py_DECREF(result);
     }
 
@@ -346,7 +355,7 @@ rpmts_InitDB(rpmtsObject * s)
 {
     int rc;
 
-    rc = rpmtsInitDB(s->ts, O_RDONLY);
+    rc = rpmtsInitDB(s->ts, 0644);
     if (rc == 0)
 	rc = rpmtsCloseDB(s->ts);
 
@@ -397,19 +406,27 @@ rpmts_HdrFromFdno(rpmtsObject * s, PyObject *arg)
 {
     PyObject *ho = NULL;
     rpmfdObject *fdo = NULL;
-    Header h;
+    Header h = NULL;
     rpmRC rpmrc;
+    PyObject *fdo_source;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
 
-    if (!PyArg_Parse(arg, "O&:HdrFromFdno", rpmfdFromPyObject, &fdo))
+    if (!PyArg_Parse(arg, "O:HdrFromFdno", &fdo_source))
     	return NULL;
+    if(!rpmfdFromPyObject(modstate, fdo_source, &fdo)) {
+	return NULL;
+    }
 
     Py_BEGIN_ALLOW_THREADS;
     rpmrc = rpmReadPackageFile(s->ts, rpmfdGetFd(fdo), NULL, &h);
     Py_END_ALLOW_THREADS;
     Py_XDECREF(fdo);
 
-    if (rpmrc == RPMRC_OK) {
-	ho = hdr_Wrap(&hdr_Type, h);
+    if (h) {
+	ho = hdr_Wrap(modstate->hdr_Type, h);
     } else {
 	Py_INCREF(Py_None);
 	ho = Py_None;
@@ -439,26 +456,6 @@ rpmts_HdrCheck(rpmtsObject * s, PyObject *obj)
     return Py_BuildValue("(iN)", rpmrc, utf8FromString(msg));
 }
 
-static PyObject *
-rpmts_PgpPrtPkts(rpmtsObject * s, PyObject * args, PyObject * kwds)
-{
-    PyObject * blob;
-    unsigned char * pkt;
-    unsigned int pktlen;
-    int rc;
-    char * kwlist[] = {"octets", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "S:PgpPrtPkts", kwlist, &blob))
-    	return NULL;
-
-    pkt = (unsigned char *)PyBytes_AsString(blob);
-    pktlen = PyBytes_Size(blob);
-
-    rc = pgpPrtPkts(pkt, pktlen, NULL, 1);
-
-    return Py_BuildValue("i", rc);
-}
-
 static PyObject *
 rpmts_PgpImportPubkey(rpmtsObject * s, PyObject * args, PyObject * kwds)
 {
@@ -482,8 +479,12 @@ rpmts_PgpImportPubkey(rpmtsObject * s, PyObject * args, PyObject * kwds)
 
 static PyObject *rpmts_setKeyring(rpmtsObject *s, PyObject *arg)
 {
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
     rpmKeyring keyring = NULL;
-    if (arg == Py_None || rpmKeyringFromPyObject(arg, &keyring)) {
+    if (arg == Py_None || rpmKeyringFromPyObject(modstate, arg, &keyring)) {
 	return PyBool_FromLong(rpmtsSetKeyring(s->ts, keyring) == 0);
     } else {
 	PyErr_SetString(PyExc_TypeError, "rpm.keyring or None expected");
@@ -493,6 +494,10 @@ static PyObject *rpmts_setKeyring(rpmtsObject *s, PyObject *arg)
 
 static PyObject *rpmts_getKeyring(rpmtsObject *s, PyObject *args, PyObject *kwds)
 {
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
     rpmKeyring keyring = NULL;
     int autoload = 1;
     char * kwlist[] = { "autoload", NULL };
@@ -503,18 +508,17 @@ static PyObject *rpmts_getKeyring(rpmtsObject *s, PyObject *args, PyObject *kwds
 
     keyring = rpmtsGetKeyring(s->ts, autoload);
     if (keyring) {
-	return rpmKeyring_Wrap(&rpmKeyring_Type, keyring);
+	return rpmKeyring_Wrap(modstate->rpmKeyring_Type, keyring);
     } else {
 	Py_RETURN_NONE;
     }
 }
 
 static void *
-rpmtsCallback(const void * hd, const rpmCallbackType what,
+rpmtsCallback(const void * arg, const rpmCallbackType what,
 		         const rpm_loff_t amount, const rpm_loff_t total,
 	                 const void * pkgKey, rpmCallbackData data)
 {
-    Header h = (Header) hd;
     struct rpmtsCallbackType_s * cbInfo = data;
     PyObject * pkgObj = (PyObject *) pkgKey;
     PyObject * args, * result;
@@ -524,21 +528,45 @@ rpmtsCallback(const void * hd, const rpmCallbackType what,
 
     PyEval_RestoreThread(cbInfo->_save);
 
-    /* Synthesize a python object for callback (if necessary). */
-    if (pkgObj == NULL) {
-	if (h) {
-	    pkgObj = utf8FromString(headerGetString(h, RPMTAG_NAME));
-	} else {
-	    pkgObj = Py_None;
-	    Py_INCREF(pkgObj);
-	}
-    } else
-	Py_INCREF(pkgObj);
+    if (cbInfo->tso == NULL) {
+	PyErr_SetString(PyExc_SystemError, "callback tso not set");
+	return NULL;
+    }
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)cbInfo->tso);
+    if (!modstate) {
+	return NULL;
+    }
 
-    args = Py_BuildValue("(iLLOO)", what, amount, total, pkgObj, cbInfo->data);
-    result = PyEval_CallObject(cbInfo->cb, args);
+    if (cbInfo->style == 0) {
+	/* Synthesize a python object for callback (if necessary). */
+	if (pkgObj == NULL) {
+	    if (arg) {
+		Header h = (Header) arg;
+		pkgObj = utf8FromString(headerGetString(h, RPMTAG_NAME));
+	    } else {
+		pkgObj = Py_None;
+		Py_INCREF(pkgObj);
+	    }
+	} else
+	    Py_INCREF(pkgObj);
+
+	args = Py_BuildValue("(iLLOO)", what, amount, total,
+				pkgObj, cbInfo->data);
+	Py_DECREF(pkgObj);
+    } else {
+	PyObject *o;
+	if (arg) {
+	    o = rpmte_Wrap(modstate->rpmte_Type, (rpmte) arg);
+	} else {
+	    o = Py_None;
+	    Py_INCREF(o);
+	}
+	args = Py_BuildValue("(OiLLO)", o, what, amount, total, cbInfo->data);
+	Py_DECREF(o);
+    }
+
+    result = PyObject_Call(cbInfo->cb, args, NULL);
     Py_DECREF(args);
-    Py_DECREF(pkgObj);
 
     if (!result) {
 	die(cbInfo->cb);
@@ -571,8 +599,12 @@ rpmtsCallback(const void * hd, const rpmCallbackType what,
 static PyObject *
 rpmts_Problems(rpmtsObject * s)
 {
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
     rpmps ps = rpmtsProblems(s->ts);
-    PyObject *problems = rpmps_AsList(ps);
+    PyObject *problems = rpmps_AsList(modstate, ps);
     rpmpsFree(ps);
     return problems;
 }
@@ -590,6 +622,7 @@ rpmts_Run(rpmtsObject * s, PyObject * args, PyObject * kwds)
 	return NULL;
 
     cbInfo.tso = s;
+    cbInfo.style = rpmtsGetNotifyStyle(s->ts);
     cbInfo._save = PyEval_SaveThread();
 
     if (cbInfo.cb != NULL) {
@@ -615,6 +648,10 @@ rpmts_iternext(rpmtsObject * s)
 {
     PyObject * result = NULL;
     rpmte te;
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
 
     /* Reset iterator on 1st entry. */
     if (s->tsi == NULL) {
@@ -625,7 +662,7 @@ rpmts_iternext(rpmtsObject * s)
 
     te = rpmtsiNext(s->tsi, 0);
     if (te != NULL) {
-	result = rpmte_Wrap(&rpmte_Type, te);
+	result = rpmte_Wrap(modstate->rpmte_Type, te);
     } else {
 	s->tsi = rpmtsiFree(s->tsi);
     }
@@ -645,17 +682,17 @@ rpmts_Match(rpmtsObject * s, PyObject * args, PyObject * kwds)
     int len = 0;
     rpmDbiTagVal tag = RPMDBI_PACKAGES;
     char * kwlist[] = {"tagNumber", "key", NULL};
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&O:Match", kwlist,
 	    tagNumFromPyObject, &tag, &Key))
 	return NULL;
 
     if (Key) {
-	if (PyInt_Check(Key)) {
-	    lkey = PyInt_AsLong(Key);
-	    key = (char *)&lkey;
-	    len = sizeof(lkey);
-	} else if (PyLong_Check(Key)) {
+	if (PyLong_Check(Key)) {
 	    lkey = PyLong_AsLong(Key);
 	    key = (char *)&lkey;
 	    len = sizeof(lkey);
@@ -675,12 +712,13 @@ rpmts_Match(rpmtsObject * s, PyObject * args, PyObject * kwds)
     if (rpmtsGetRdb(s->ts) == NULL) {
 	int rc = rpmtsOpenDB(s->ts, O_RDONLY);
 	if (rc || rpmtsGetRdb(s->ts) == NULL) {
-	    PyErr_SetString(pyrpmError, "rpmdb open failed");
+	    PyErr_SetString(modstate->pyrpmError, "rpmdb open failed");
 	    goto exit;
 	}
     }
 
-    mio = rpmmi_Wrap(&rpmmi_Type, rpmtsInitIterator(s->ts, tag, key, len), (PyObject*)s);
+    mio = rpmmi_Wrap(modstate->rpmmi_Type,
+                     rpmtsInitIterator(s->ts, tag, key, len), (PyObject*)s);
 
 exit:
     Py_XDECREF(str);
@@ -692,6 +730,10 @@ rpmts_index(rpmtsObject * s, PyObject * args, PyObject * kwds)
     rpmDbiTagVal tag;
     PyObject *mio = NULL;
     char * kwlist[] = {"tag", NULL};
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	    return NULL;
+    }
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&:Keys", kwlist,
               tagNumFromPyObject, &tag))
@@ -701,7 +743,7 @@ rpmts_index(rpmtsObject * s, PyObject * args, PyObject * kwds)
     if (rpmtsGetRdb(s->ts) == NULL) {
 	int rc = rpmtsOpenDB(s->ts, O_RDONLY);
 	if (rc || rpmtsGetRdb(s->ts) == NULL) {
-	    PyErr_SetString(pyrpmError, "rpmdb open failed");
+	    PyErr_SetString(modstate->pyrpmError, "rpmdb open failed");
 	    goto exit;
 	}
     }
@@ -711,7 +753,7 @@ rpmts_index(rpmtsObject * s, PyObject * args, PyObject * kwds)
         PyErr_SetString(PyExc_KeyError, "No index for this tag");
         return NULL;
     }
-    mio = rpmii_Wrap(&rpmii_Type, ii, (PyObject*)s);
+    mio = rpmii_Wrap(modstate->rpmii_Type, ii, (PyObject*)s);
 
 exit:
     return mio;
@@ -727,6 +769,8 @@ static struct PyMethodDef rpmts_methods[] = {
   "  mode : optional argument that specifies if this package should be\n\t\tinstalled ('i'), upgraded ('u')"},
  {"addReinstall",	(PyCFunction) rpmts_AddReinstall,	METH_VARARGS,
   "ts.addReinstall(hdr, data) -- Adds transaction elements\nrepresenting a reinstall of an already installed package.\n\nSee addInstall for details."},
+ {"addRestore",	(PyCFunction) rpmts_AddRestore,			METH_VARARGS,
+  "ts.addRestore(hdr) -- Adds transaction elements\nrepresenting a restore of an already installed package."},
  {"addErase",	(PyCFunction) rpmts_AddErase,	METH_VARARGS|METH_KEYWORDS,
   "addErase(name) -- Add a transaction element representing an erase\nof an installed package.\n\n"
   "  name: the package name to be erased"},
@@ -808,8 +852,6 @@ Remove all elements from the transaction set\n" },
   "the digest or signature is verified.\n\n"
   "\thdrblob : unloaded header blob\n"
   "Return tuple (int status, message string)"},
- {"pgpPrtPkts",	(PyCFunction) rpmts_PgpPrtPkts,	METH_VARARGS|METH_KEYWORDS,
-  "pgpPrtPkts(octets) -- Print/parse a OpenPGP packet(s).\n\nReturn 0 on success." },
  {"pgpImportPubkey",	(PyCFunction) rpmts_PgpImportPubkey,	METH_VARARGS|METH_KEYWORDS,
   "pgpImportPubkey(pubkey) -- Import public key packet." },
  {"getKeyring",	(PyCFunction) rpmts_getKeyring,	METH_VARARGS|METH_KEYWORDS, 
@@ -831,16 +873,31 @@ Remove all elements from the transaction set\n" },
 
 static void rpmts_dealloc(rpmtsObject * s)
 {
+    PyObject_GC_UnTrack(s);
 
     s->ts = rpmtsFree(s->ts);
     Py_XDECREF(s->scriptFd);
     Py_XDECREF(s->keyList);
-    Py_TYPE(s)->tp_free((PyObject *)s);
+    PyTypeObject *type = Py_TYPE(s);
+    freefunc free = PyType_GetSlot(type, Py_tp_free);
+    free(s);
+    Py_DECREF(type);
+}
+
+static int rpmts_traverse(rpmtsObject * s, visitproc visit, void *arg)
+{
+    Py_VISIT(s->scriptFd);
+    Py_VISIT(s->keyList);
+    if (python_version >= 0x03090000) {
+        Py_VISIT(Py_TYPE(s));
+    }
+    return 0;
 }
 
 static PyObject * rpmts_new(PyTypeObject * subtype, PyObject *args, PyObject *kwds)
 {
-    rpmtsObject * s = (rpmtsObject *)subtype->tp_alloc(subtype, 0);
+    allocfunc subtype_alloc = (allocfunc)PyType_GetSlot(subtype, Py_tp_alloc);
+    rpmtsObject *s = (rpmtsObject *)subtype_alloc(subtype, 0);
     if (s == NULL) return NULL;
 
     s->ts = rpmtsCreate();
@@ -878,11 +935,33 @@ static PyObject *rpmts_get_rootDir(rpmtsObject *s, void *closure)
     return utf8FromString(rpmtsRootDir(s->ts));
 }
 
+static PyObject *rpmts_get_cbstyle(rpmtsObject *s, void *closure)
+{
+    return Py_BuildValue("i", rpmtsGetNotifyStyle(s->ts));
+}
+
+static int rpmts_set_cbstyle(rpmtsObject *s, PyObject *value, void *closure)
+{
+    int cbstyle;
+    int rc = -1;
+    if (PyArg_Parse(value, "i", &cbstyle))
+	rc = rpmtsSetNotifyStyle(s->ts, cbstyle);
+    return rc;
+}
+
 static int rpmts_set_scriptFd(rpmtsObject *s, PyObject *value, void *closure)
 {
+    PyObject *fdo_source;
     rpmfdObject *fdo = NULL;
     int rc = 0;
-    if (PyArg_Parse(value, "O&", rpmfdFromPyObject, &fdo)) {
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return -1;
+    }
+    if (PyArg_Parse(value, "O", &fdo_source)) {
+	if(!rpmfdFromPyObject(modstate, fdo_source, &fdo)) {
+	    return -1;
+	}
 	Py_XDECREF(s->scriptFd);
 	s->scriptFd = fdo;
 	rpmtsSetScriptFd(s->ts, rpmfdGetFd(s->scriptFd));
@@ -1010,6 +1089,8 @@ static char rpmts_doc[] =
   "The transaction set offers an read only iterable interface for the\ntransaction elements added by the .addInstall(), .addErase() and\n.addReinstall() methods.";
 
 static PyGetSetDef rpmts_getseters[] = {
+	{"cbStyle",	(getter)rpmts_get_cbstyle, (setter)rpmts_set_cbstyle,
+	 "progress callback style" },
 	/* only provide a setter until we have rpmfd wrappings */
 	{"scriptFd",	NULL,	(setter)rpmts_set_scriptFd,
 	 "write only, file descriptor the output of script gets written to." },
@@ -1026,45 +1107,23 @@ static PyGetSetDef rpmts_getseters[] = {
 	{ NULL }
 };
 
-PyTypeObject rpmts_Type = {
-	PyVarObject_HEAD_INIT(&PyType_Type, 0)
-	"rpm.ts",			/* tp_name */
-	sizeof(rpmtsObject),		/* tp_size */
-	0,				/* tp_itemsize */
-	(destructor) rpmts_dealloc, 	/* tp_dealloc */
-	0,				/* tp_print */
-	(getattrfunc)0, 		/* tp_getattr */
-	(setattrfunc)0,			/* tp_setattr */
-	0,				/* tp_compare */
-	0,				/* tp_repr */
-	0,				/* tp_as_number */
-	0,				/* tp_as_sequence */
-	0,				/* tp_as_mapping */
-	0,				/* tp_hash */
-	0,				/* tp_call */
-	0,				/* tp_str */
-	PyObject_GenericGetAttr, 	/* tp_getattro */
-	PyObject_GenericSetAttr,	/* tp_setattro */
-	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,	/* tp_flags */
-	rpmts_doc,			/* tp_doc */
-	0,				/* tp_traverse */
-	0,				/* tp_clear */
-	0,				/* tp_richcompare */
-	0,				/* tp_weaklistoffset */
-	PyObject_SelfIter,		/* tp_iter */
-	(iternextfunc) rpmts_iternext,	/* tp_iternext */
-	rpmts_methods,			/* tp_methods */
-	0,				/* tp_members */
-	rpmts_getseters,		/* tp_getset */
-	0,				/* tp_base */
-	0,				/* tp_dict */
-	0,				/* tp_descr_get */
-	0,				/* tp_descr_set */
-	0,				/* tp_dictoffset */
-	(initproc) rpmts_init,		/* tp_init */
-	0,				/* tp_alloc */
-	(newfunc) rpmts_new,		/* tp_new */
-	0,				/* tp_free */
-	0,				/* tp_is_gc */
+static PyType_Slot rpmts_Type_Slots[] = {
+    {Py_tp_dealloc, rpmts_dealloc},
+    {Py_tp_traverse, rpmts_traverse},
+    {Py_tp_getattro, PyObject_GenericGetAttr},
+    {Py_tp_setattro, PyObject_GenericSetAttr},
+    {Py_tp_doc, rpmts_doc},
+    {Py_tp_iter, PyObject_SelfIter},
+    {Py_tp_iternext, rpmts_iternext},
+    {Py_tp_methods, rpmts_methods},
+    {Py_tp_getset, rpmts_getseters},
+    {Py_tp_init, rpmts_init},
+    {Py_tp_new, rpmts_new},
+    {0, NULL},
+};
+PyType_Spec rpmts_Type_Spec = {
+    .name = "rpm.ts",
+    .basicsize = sizeof(rpmtsObject),
+    .flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_IMMUTABLETYPE,
+    .slots = rpmts_Type_Slots,
 };
